#!/usr/bin/env python
#
#  BLIS
#  An object-based framework for developing high-performance BLAS-like
#  libraries.
#
#  Copyright (C) 2014, The University of Texas at Austin
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are
#  met:
#   - Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   - Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   - Neither the name(s) of the copyright holder(s) nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#  HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#

# Import modules
import os
import sys
import getopt
import re

def print_usage():

	my_print( " " )
	my_print( " %s" % script_name )
	my_print( " " )
	my_print( " Field G. Van Zee" )
	my_print( " " )
	my_print( " Generate a monolithic header by recursively replacing all #include" )
	my_print( " directives in a selected file with the contents of the header files" )
	my_print( " they reference." )
	my_print( " " )
	my_print( " Usage:" )
	my_print( " " )
	my_print( "   %s header header_out temp_dir dir_list" % script_name )
	my_print( " " )
	my_print( " Arguments:" )
	my_print( " " )
	my_print( "   header        The filepath to the top-level header, which is the file" )
	my_print( "                 that will #include all other header files." )
	my_print( " " )
	my_print( "   header_out    The filepath of the file into which the script will output" )
	my_print( "                 the monolithic header." )
	my_print( " " )
	my_print( "   temp_dir      A directory in which temporary files may be created." )
	my_print( "                 NOTE: No temporary files are created in the current" )
	my_print( "                 implementation, but this argument must still be specified." )
	my_print( " " )
	my_print( "   dir_list      The list of directory paths in which to search for the" )
	my_print( "                 headers that are #included by 'header'. By default, these" )
	my_print( "                 directories are scanned for .h files, but sub-directories" )
	my_print( "                 within the various directories are not inspected. If the" )
	my_print( "                 -r option is given, these directories are recursively" )
	my_print( "                 scanned. In either case, the subset of directories scanned" )
	my_print( "                 that actually contains .h files is then searched whenever" )
	my_print( "                 a #include directive is encountered in 'header' (or any" )
	my_print( "                 file subsequently #included). If a referenced header file" )
	my_print( "                 is not found, the #include directive is left untouched and" )
	my_print( "                 translated directly into 'header_out'." )
	my_print( " " )
	my_print( " The following options are accepted:" )
	my_print( " " )
	my_print( "   -r          recursive" )
	my_print( "                 Scan the directories listed in 'dir_list' recursively when" )
	my_print( "                 searching for .h header files. By default, the directories" )
	my_print( "                 are not searched recursively." )
	my_print( " " )
	my_print( "   -c          strip C-style comments" )
	my_print( "                 Strip comments enclosed in /* */ delimiters from the" )
	my_print( "                 output, including multi-line comments. By default, C-style" )
	my_print( "                 comments are not stripped." )
	my_print( " " )
	my_print( "   -o SCRIPT   output script name" )
	my_print( "                 Use SCRIPT as a prefix when outputting messages instead" )
	my_print( "                 the script's actual name. Useful when the current script" )
	my_print( "                 is going to be called from within another, higher-level" )
	my_print( "                 driver script and seeing the current script's name might" )
	my_print( "                 unnecessarily confuse the user." )
	my_print( " " )
	my_print( "   -v [0|1|2]  verboseness level" )
	my_print( "                 level 0: silent  (no output)" )
	my_print( "                 level 1: default (single character '.' per header)" )
	my_print( "                 level 2: verbose (several lines per header)." )
	my_print( " " )
	my_print( "   -h          help" )
	my_print( "                 Output this information and exit." )
	my_print( " " )


# ------------------------------------------------------------------------------

def canonicalize_ws( s ):

	return re.sub( r'\s+', ' ', s ).strip()

# ---

def my_print( s ):

	sys.stdout.write( "%s\n" % s )

# ---

#def echov1( s ):
#
#	if verbose_flag == "1":
#		print "%s: %s" % ( output_name, s )

def echov1_n( s ):

	if verbose_flag == "1":
		sys.stdout.write( s )
		sys.stdout.flush()

def echov1_n2( s ):

	if verbose_flag == "1":
		sys.stdout.write( "%s\n" % s )
		sys.stdout.flush()

# ---

def echov2( s ):

	if verbose_flag == "2":
		sys.stdout.write( "%s: %s\n" % ( output_name, s ) )
		sys.stdout.flush()

def echov2_n( s ):

	if verbose_flag == "2":
		sys.stdout.write( output_name )
		sys.stdout.write( ": " )
		sys.stdout.write( s )
		sys.stdout.flush()

def echov2_n2( s ):

	if verbose_flag == "2":
		sys.stdout.write( "%s\n" % s )
		sys.stdout.flush()

# ------------------------------------------------------------------------------

def list_contains_header( items ):

	rval = False
	for item in items:

		is_h = re.search( r"\.h", item )

		if is_h:
			rval = True
			break

	return rval

# ------------------------------------------------------------------------------

def get_header_path( filename, header_dirpaths ):

	filepath = None

	# Search each directory path for the filename given.
	for dirpath in header_dirpaths:

		# Construct a possible path to the sought-after file.
		cur_filepath = "%s/%s" % ( dirpath, filename )

		# Check whether the file exists.
		found = os.path.exists( cur_filepath )
		if found:
			filepath = cur_filepath
			break

	return filepath

# ------------------------------------------------------------------------------

def strip_cstyle_comments( string ):

	return re.sub( r"/\*.*?\*/", "", string, flags=re.S )

# ------------------------------------------------------------------------------

def flatten_header( inputfile, header_dirpaths, cursp ):

	# This string is inserted after #include directives after having
	# determined that they are not present in the directory tree.
	skipstr  = "// skipped"
	beginstr = "// begin "
	endstr   = "// end "

	ostring  = ""

	# Open the input file to process.
	ifile = open( inputfile, "r" )

	# A counter to track the line number being parsed within the current file.
	# This counter, when selectively encoded into the flattened header via #line
	# directives, facilitates easier debugging. (When the compiler finds an
	# issue, it will be able to refer to the line number within the constituent
	# header file rather than the flattened one.)
	lineno = 0

	# Iterate over the lines in the file.
	while True:

		# Increment the line number.
		lineno += 1

		# Read a line in the file.
		line = ifile.readline()

		# Check for EOF.
		if line == '': break

		# Check for the #include directive and isolate the header name within
		# a group (parentheses).
		#result = re.search( '^[\s]*#include (["<])([\w\.\-/]*)([">])', line )
		result = regex.search( line )

		# If the line contained a #include directive, we must try to replace
		# it with the contents of the header referenced by the directive.
		if result:

			# Extract the header file referenced in the #include directive,
			# saved as the second group in the regular expression
			# above.
			header = result.group(2)

			echov2( "%sfound reference to '%s'." % ( cursp, header ) )

			# Search for the path to the header referenced in the #include
			# directive.
			header_path = get_header_path( header, header_dirpaths )

			# First, check if the header is our root header (and if so, ignore it).
			# Otherwise, if the header was found, we recurse. Otherwise, we output
			# the #include directive with a comment indicating that it as skipped
			if header == root_inputfile:

				markl = result.group(1)
				markr = result.group(3)

				echov2( "%sthis is the root header '%s'; commenting out / skipping." \
				        % ( cursp, header ) )

				# If the header found is our root header, then we cannot
				# recurse into it lest we enter an infinite loop. Output the
				# line but make sure it's commented out entirely.
				ostring += "%s #include %c%s%c %c" \
				           % ( skipstr, markl, header, markr, '\n' )

			elif header_path:

				echov2( "%slocated file '%s'; recursing." \
				        % ( cursp, header_path ) )

				# Mark the beginning of the header being inserted.
				ostring += "%s%s%c" % ( beginstr, header, '\n' )
				if line_numbers:
					ostring += "#line %d \"%s\"%c\n" % ( 1, header_path, '\n' )

				# Recurse on the header, accumulating the string.
				ostring += flatten_header( header_path, header_dirpaths, cursp + "  " )

				# Mark the end of the header being inserted.
				ostring += "%s%s%c" % ( endstr, header, '\n' )
				if line_numbers:
					ostring += "#line %d \"%s\"%c\n" % ( lineno+1, inputfile, '\n' )

				echov2( "%sheader file '%s' fully processed." \
				        % ( cursp, header_path ) )

			else:

				markl = result.group(1)
				markr = result.group(3)

				echov2( "%scould not locate file '%s'; marking as skipped." \
				        % ( cursp, header ) )

				# If the header was not found, output the line with a
				# comment that the header was skipped.
				ostring += "#include %c%s%c %s%c" \
				           % ( markl, header, markr, skipstr, '\n' )
			# endif

		else:
			# If the line did not contain a #include directive, simply output
			# the line verbatim.
			ostring += "%s" % line

		# endif

	# endwhile

	# Close the input file.
	ifile.close()

	echov1_n( "." )

	return ostring

# ------------------------------------------------------------------------------

def find_header_dirs( dirpath ):

	header_dirpaths = []
	for root, dirs, files in os.walk( dirpath, topdown=True ):

		echov2_n( "scanning contents of %s" % root )

		if list_contains_header( files ):

			echov2_n2( "...found headers" )
			header_dirpaths.append( root )

		else:
			echov2_n2( "" )

		#endif

	#endfor

	return header_dirpaths

# ------------------------------------------------------------------------------

# Global variables.
script_name    = None
output_name    = None
strip_comments = None
recursive_flag = None
line_numbers   = None
verbose_flag   = None
regex          = None
root_inputfile = None

def main():

	global script_name
	global output_name
	global strip_comments
	global recursive_flag
	global line_numbers
	global verbose_flag
	global regex
	global root_inputfile

	# Obtain the script name.
	path, script_name = os.path.split(sys.argv[0])

	output_name    = script_name

	strip_comments = False
	recursive_flag = False
	line_numbers   = False
	verbose_flag   = "1"

	nestsp         = "  "

	# Process our command line options.
	try:
		opts, args = getopt.getopt( sys.argv[1:], "o:rclhv:" )

	except getopt.GetoptError as err:
		# print help information and exit:
		my_print( str(err) ) # will print something like "option -a not recognized"
		print_usage()
		sys.exit(2)

	for opt, optarg in opts:
		if   opt == "-o":
			output_name = optarg
		elif opt == "-r":
			recursive_flag = True
		elif opt == "-l":
			line_numbers = True
		elif opt == "-c":
			strip_comments = True
		elif opt == "-v":
			verbose_flag = optarg
		elif opt == "-h":
			print_usage()
			sys.exit()
		else:
			print_usage()
			sys.exit()

	if line_numbers and strip_comments:
		my_print( "WARNING: stripping comments will result in inaccurate line numbers" )

	# Make sure that the verboseness level is valid.
	if ( verbose_flag != "0" and
	     verbose_flag != "1" and
	     verbose_flag != "2" ):
		my_print( "%s Invalid verboseness argument: %s" \
		                  % output_name, verbose_flag )
		sys.exit()

	# Print usage if we don't have exactly four arguments.
	if len( args ) != 4:
		print_usage()
		sys.exit()

	# Acquire the four required arguments:
	# - the input header file,
	# - the output header file,
	# - the temporary directory in which we can write intermediate files,
	# - the list of directories in which to search for the headers.
	inputfile  = args[0]
	outputfile = args[1]
	temp_dir   = args[2]
	dir_list   = args[3]

	# Save the filename (basename) part of the input file (or root file) into a
	# global variable that we can access later from within flatten_header().
	root_inputfile = os.path.basename( inputfile )

	# Separate the directories into distinct strings.
	dir_list = dir_list.split()

	# First, confirm that the directories in dir_list are valid.
	dir_list_checked = []
	for item in dir_list:

		#absitem = os.path.abspath( item )

		echov2_n( "checking " + item )

		if os.path.exists( item ):
			dir_list_checked.append( item )
			echov2_n2( "...directory exists." )
		else:
			echov2_n2( "...invalid directory; omitting." )

	# endfor

	# Overwrite the original dir_list with the updated copy that omits
	# invalid directories.
	dir_list = dir_list_checked

	echov2( "check summary:" )
	echov2( "  accessible directories:" )
	echov2( "  %s" % ' '.join( dir_list ) )

	# Generate a list of directories (header_dirpaths) which will be searched
	# whenever a #include directive is encountered. The method by which
	# header_dirpaths is compiled will depend on whether the recursive flag
	# was given.
	if recursive_flag:

		header_dirpaths = []
		for d in dir_list:

			# For each directory in dir_list, recursively walk that directory
			# and return a list of directories that contain headers.
			d_dirpaths = find_header_dirs( d )

			# Add the list resulting from the current search to the running
			# list of directory paths that contain headers.
			header_dirpaths += d_dirpaths

		# endfor

	else:

		# If the recursive flag was not given, we can just use dir_list
		# as-is, though we opt to filter out the directories that don't
		# contain .h files.

		header_dirpaths = []
		for d in dir_list:

			echov2_n( "scanning %s" % d )

			# Acquire a list of the directory's contents.
			sub_items = os.listdir( d )

			# If there is at least one header present, add the current
			# directory to the list of header directories.
			if list_contains_header( sub_items ):
				header_dirpaths.append( d )
				echov2_n2( "...found headers." )
			else:
				echov2_n2( "...no headers found." )
			# endif

		# endfor

	# endfor

	echov2( "scan summary:" )
	echov2( "  headers found in:" )
	echov2( "  %s" % ' '.join( header_dirpaths ) )

	echov2( "preparing to monolithify '%s'" % inputfile )

	echov2( "new header will be saved to '%s'" % outputfile )

	echov1_n( "." )

	# Open the output file.
	ofile = open( outputfile, "w" )

	# Precompile the main regular expression used to isolate #include
	# directives and the headers they reference. This regex object will
	# get reused over and over again in flatten_header().
	regex = re.compile( r'^[\s]*#include (["<])([\w\.\-/]*)([">])' )

	# Recursively substitute headers for occurrences of #include directives.
	final_string = flatten_header( inputfile, header_dirpaths, nestsp )

	# Strip C-style comments from the final output, if requested.
	if strip_comments:
		final_string = strip_cstyle_comments( final_string )

	# Write the lines to the file.
	ofile.write( final_string )

	# Close the output file.
	ofile.close()

	echov2( "substitution complete." )
	echov2( "monolithic header saved as '%s'" % outputfile )

	echov1_n2( "." )

	return 0




if __name__ == "__main__":
	main()



#!/usr/bin/env python3
#
#  BLIS    
#  An object-based framework for developing high-performance BLAS-like
#  libraries.
#
#  Copyright (C) 2018, The University of Texas at Austin
#  Copyright (C) 2018 - 2019, Advanced Micro Devices, Inc.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are
#  met:
#   - Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   - Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   - Neither the name(s) of the copyright holder(s) nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#  HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#

# Import modules
import os
import sys
import getopt
import re
import subprocess
import time
import statistics


def print_usage():

	my_print( " " )
	my_print( " %s" % script_name )
	my_print( " " )
	my_print( " Field G. Van Zee" )
	my_print( " " )
	my_print( " Repeatedly run a test driver and accumulate statistics for the" )
	my_print( " output." )
	my_print( " " )
	my_print( " Usage:" )
	my_print( " " )
	my_print( "   %s [options] drivername" % script_name )
	my_print( " " )
	my_print( " Arguments:" )
	my_print( " " )
	my_print( "   drivername    The filename/path of the test driver to run. The" )
	my_print( "                 test driver must output its performance data to" )
	my_print( "                 standard output." )
	my_print( " " )
	my_print( " The following options are accepted:" )
	my_print( " " )
	my_print( "   -c num      performance column index" )
	my_print( "                 Find the performance result in column index <num> of" )
	my_print( "                 the test driver's output. Here, a column is defined" )
	my_print( "                 as a contiguous sequence of non-whitespace characters," )
	my_print( "                 with the column indices beginning at 0. By default," )
	my_print( "                 the second-to-last column index in the output is used." )
	my_print( " " )
	my_print( "   -d delay    sleep() delay" )
	my_print( "                 Wait <delay> seconds after each execution of the" )
	my_print( "                 test driver. The default delay is 0." )
	my_print( " " )
	my_print( "   -n niter    number of iterations" )
	my_print( "                 Execute the test driver <niter> times. The default" )
	my_print( "                 value is 10." )
	my_print( " " )
	my_print( "   -q          quiet; summary only" )
	my_print( "                 Do not output statistics after every new execution of" )
	my_print( "                 the test driver; instead, only output the final values" )
	my_print( "                 after all iterations are complete. The default is to" )
	my_print( "                 output updated statistics after each iteration." )
	my_print( " " )
	my_print( "   -h          help" )
	my_print( "                 Output this information and exit." )
	my_print( " " )


# ------------------------------------------------------------------------------

def my_print( s ):

	sys.stdout.write( "%s\n" % s )
	#sys.stdout.flush()

# ------------------------------------------------------------------------------

# Global variables.
script_name    = None
output_name    = None

def main():

	global script_name
	global output_name

	# Obtain the script name.
	path, script_name = os.path.split(sys.argv[0])

	output_name = script_name

	# Default values for optional arguments.
	#perf_col = 9
	perf_col = -1
	delay    = 0
	niter    = 10
	quiet    = False

	# Process our command line options.
	try:
		opts, args = getopt.getopt( sys.argv[1:], "c:d:n:hq" )

	except getopt.GetoptError as err:
		# print help information and exit:
		my_print( str(err) ) # will print something like "option -a not recognized"
		print_usage()
		sys.exit(2)

	for opt, optarg in opts:
		if   opt == "-c":
			perf_col = optarg
		elif opt == "-d":
			delay = optarg
		elif opt == "-n":
			niter = optarg
		elif opt == "-q":
			quiet = True
		elif opt == "-h":
			print_usage()
			sys.exit()
		else:
			print_usage()
			sys.exit()

	# Print usage if we don't have exactly one argument.
	if len( args ) != 1:
		print_usage()
		sys.exit()

	# Acquire our only mandatory argument: the name of the test driver.
	driverfile = args[0]

	#my_print( "test driver: %s" % driverfile )
	#my_print( "column num:  %s" % perf_col )
	#my_print( "delay:       %s" % delay )
	#my_print( "num iter:    %s" % niter )

	# Build a list of iterations.
	iters = range( int(niter) )

	# Run the test driver once to detect the number of lines of output.
	p = subprocess.run( driverfile, stdout=subprocess.PIPE )
	lines0 = p.stdout.decode().splitlines()
	num_lines0 = int(len(lines0))

	# Initialize the list of lists (one list per performance result).
	aperf = []
	for i in range( num_lines0 ):
		aperf.append( [] )

	for it in iters:

		# Run the test driver.
		p = subprocess.run( driverfile, stdout=subprocess.PIPE )

		# Acquire the lines of output.
		lines = p.stdout.decode().splitlines()

		# Accumulate the test driver's latest results into aperf.
		for i in range( num_lines0 ):

			# Parse the current line to find the performance value.
			line  = lines[i]
			words = line.split()
			if perf_col == -1:
				perf  = words[ len(words)-2 ]
			else:
				perf  = words[ int(perf_col) ]

			# As unlikely as it is, guard against Inf and NaN.
			if float(perf) ==  float('Inf') or \
			   float(perf) == -float('Inf') or \
			   float(perf) ==  float('NaN'): perf = 0.0

			# Add the performance value to the list at the ith entry of aperf.
			aperf[i].append( float(perf) )

			# Compute stats for the current line.
			avgp = statistics.mean( aperf[i] )
			maxp =             max( aperf[i] )
			minp =             min( aperf[i] )

			# Only compute stdev() when we have two or more data points.
			if len( aperf[i] ) > 1: stdp = statistics.stdev( aperf[i] )
			else:                   stdp = 0.0

			# Construct a string to match the performance value and then
			# use that string to search-and-replace with four format specs
			# for the min, avg, max, and stdev values computed above.
			search = '%8s' % perf
			newline = re.sub( str(search), ' %7.2f %7.2f %7.2f %6.2f', line )

			# Search for the column index range that would be present if this were
			# matlab-compatible output. The index range will typically be 1:n,
			# where n is the number of columns of data.
			found_index = False
			for word in words:
				if re.match( '1:', word ):
					index_str = word
					found_index = True
					break

			# If we find the column index range, we need to update it to reflect
			# the replacement of one column of data with four, for a net increase
			# of columns. We do so via another instance of re.sub() in which we
			# search for the old index string and replace it with the new one.
			if found_index:
				last_col = int(index_str[2]) + 3
				new_index_str = '1:%1s' % last_col
				newline = re.sub( index_str, new_index_str, newline )

			# If the quiet flag was not give, output the intermediate results.
			if not quiet:
				print( newline % ( float(minp), float(avgp), float(maxp), float(stdp) ) )

		# Flush stdout after each set of output prior to sleeping.
		sys.stdout.flush()

		# Sleep for a bit until the next iteration.
		time.sleep( int(delay) )

	# If the quiet flag was given, output the final results.
	if quiet:

		for i in range( num_lines0 ):

			# Parse the current line to find the performance value (only
			# needed for call to re.sub() below).
			line  = lines0[i]
			words = line.split()
			if perf_col == -1:
				perf  = words[ len(words)-2 ]
			else:
				perf  = words[ int(perf_col) ]

			# Compute stats for the current line.
			avgp = statistics.mean( aperf[i] )
			maxp =             max( aperf[i] )
			minp =             min( aperf[i] )

			# Only compute stdev() when we have two or more data points.
			if len( aperf[i] ) > 1: stdp = statistics.stdev( aperf[i] )
			else:                   stdp = 0.0

			# Construct a string to match the performance value and then
			# use that string to search-and-replace with four format specs
			# for the min, avg, max, and stdev values computed above.
			search = '%8s' % perf
			newline = re.sub( str(search), ' %7.2f %7.2f %7.2f %6.2f', line )

			# Search for the column index range that would be present if this were
			# matlab-compatible output. The index range will typically be 1:n,
			# where n is the number of columns of data.
			found_index = False
			for word in words:
				if re.match( '1:', word ):
					index_str = word
					found_index = True
					break

			# If we find the column index range, we need to update it to reflect
			# the replacement of one column of data with four, for a net increase
			# of columns. We do so via another instance of re.sub() in which we
			# search for the old index string and replace it with the new one.
			if found_index:
				last_col = int(index_str[2]) + 3
				new_index_str = '1:%1s' % last_col
				newline = re.sub( index_str, new_index_str, newline )

			# Output the results for the current line.
			print( newline % ( float(minp), float(avgp), float(maxp), float(stdp) ) )

		# Flush stdout afterwards.
		sys.stdout.flush()


	# Return from main().
	return 0




if __name__ == "__main__":
	main()



#!/usr/bin/env python3
#
#  BLIS
#  An object-based framework for developing high-performance BLAS-like
#  libraries.
#
#  Copyright (C) 2018, The University of Texas at Austin
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are
#  met:
#   - Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   - Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   - Neither the name(s) of the copyright holder(s) nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#  HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#

# Import modules
import os
import sys
import getopt
import re
import subprocess
import datetime


def print_usage():

	my_print( " " )
	my_print( " %s" % script_name )
	my_print( " " )
	my_print( " Field G. Van Zee" )
	my_print( " " )
	my_print( " Update copyright lines of all created or modified source files currently" )
	my_print( " staged in the git index, and also insert new copyright lines where they" )
	my_print( " currently are missing. This script targets copyright lines for one" )
	my_print( " organization at a time." )
	my_print( " " )
	my_print( " Usage:" )
	my_print( " " )
	my_print( "   %s [options]" % script_name )
	my_print( " " )
	my_print( " Arguments:" )
	my_print( " " )
	my_print( " " )
	my_print( " The following options are accepted:" )
	my_print( " " )
	my_print( "   -o org      organization name" )
	my_print( "                 Update and add copyrights for an organization named <org>." )
	my_print( "                 By default, <org> is 'Advanced Micro Devices, Inc.'" )
	my_print( " " )
	my_print( "   -u          update only" )
	my_print( "                 Update existing copyrights to reflect the current year," )
	my_print( "                 but do not add any additional copyright lines. With this" )
	my_print( "                 option, the script still only updates copyright lines for" )
	my_print( "                 the specified (or default) organization. The default is" )
	my_print( "                 to update but also add copyright lines where missing." )
	my_print( " " )
	my_print( "   -d          dry run" )
	my_print( "                 Go through all of the motions, but don't actually modify" )
	my_print( "                 any files. The default behavior is to not enable dry run." )
	my_print( " " )
	my_print( "   -q          quiet" )
	my_print( "                 Do not output feedback while processing each file. The" )
	my_print( "                 default behavior is to output one line of text to stdout" )
	my_print( "                 per file updated." )
	my_print( " " )
	my_print( "   -h          help" )
	my_print( "                 Output this information and exit." )
	my_print( " " )


# ------------------------------------------------------------------------------

def my_print( s ):

	sys.stdout.write( "%s\n" % s )
	#sys.stdout.flush()

def my_echo( s ):

	if not quiet:
		sys.stdout.write( "%s: %s\n" % ( output_name, s ) )
	#sys.stdout.flush()

# ------------------------------------------------------------------------------

def main():

	global script_name
	global output_name
	global quiet

	# Obtain the script name.
	path, script_name = os.path.split(sys.argv[0])

	output_name = script_name

	# Default values for optional arguments.
	the_org     = 'Advanced Micro Devices, Inc.'
	update_only = False
	dry_run     = False
	quiet       = False

	# Process our command line options.
	try:
		opts, args = getopt.getopt( sys.argv[1:], "do:uhq" )

	except getopt.GetoptError as err:
		# print help information and exit:
		my_print( str(err) ) # will print something like "option -a not recognized"
		print_usage()
		sys.exit(2)

	for opt, optarg in opts:
		if opt == "-o":
			the_org = optarg
		elif opt == "-u":
			update_only = True
		elif opt == "-d":
			dry_run = True
		elif opt == "-q":
			quiet = True
		elif opt == "-h":
			print_usage()
			sys.exit()
		else:
			print_usage()
			sys.exit()

	# Print usage if we don't have exactly zero arguments.
	if len( args ) != 0:
		print_usage()
		sys.exit()

	# Acquire our only mandatory argument.
	#driverfile = args[0]

	# Query the current year.
	the_time = datetime.datetime.now()
	cur_year = str(the_time.year)

	# We run 'git status' with --porcelain to make the output easily parseable.
	gitstatus = 'git status --porcelain'

	# Run the 'git status' command and capture the output.
	p = subprocess.run( gitstatus, stdout=subprocess.PIPE, shell=True )
	git_lines = p.stdout.decode().splitlines()
	git_num_lines = int( len( git_lines ) )

	# Consider each line of output from 'git status'
	for i in range( git_num_lines ):

		# Parse the current line to find the performance value.
		git_line  = git_lines[i]
		git_words = git_line.split()
		mod_char  = git_line[0]

		# Check the first character of the git output. We want to only update
		# files that are new ('A'), modified ('M'), or renamed ('R').
		if mod_char != 'A' and \
		   mod_char != 'M' and \
		   mod_char != 'R': continue

		# Identify the filename for the current line of 'git status' output.
		if mod_char == 'R':
			# For renamed files, we need to reference them by their new names,
			# which appear after the "->" char sequence in git_words[2].
			filename = git_words[3]
		else:
			filename = git_words[1]

		#my_echo( "-debug---- %s" % filename )

		# Start by opening the file. (We can assume it exists since it
		# was found by 'git status', so no need to check for existence.)
		# Read all lines in the file and then close it.
		f = open( filename, "r" )
		file_lines = f.readlines()
		f.close()

		# Concatenate all lines in the file into one string.
		file_string = "".join( file_lines )

		# Search for an existing copyright line.
		has_cr = re.search( r'Copyright \(C\)', file_string )

		# If the file does not have any copyright notice in it already, we
		# assume we don't need to update it.
		if not has_cr:
			my_echo( "[nocrline] %s" % filename )
			continue

		# Check whether the file already has a copyright for the_org. We may
		# need to use this information later.
		has_org_cr = re.search( r'Copyright \(C\) ([0-9][0-9][0-9][0-9]), %s' % the_org, file_string )

		# Initialize the list of processed (potentially modified) file lines.
		mod_file_lines = []

		# At this point we know that the file has at least one copyright, and
		# has_org_cr encodes whether it already has a copyright for the_org.

		# We process the files that we know already have copyrights for the_org
		# differently from the files that do not yet have them.
		if has_org_cr:

			# Iterate through the lines in the current file.
			for line in file_lines:

				result = re.search( r'Copyright \(C\) ([0-9][0-9][0-9][0-9]), %s' % the_org, line )

				# If the current line matches a copyright line for the_org...
				if result:

					# Extract the year saved as the first/only group in the
					# regular expression.
					old_year = result.group(1)

					# Don't need to update the year if it's already up-to-date.
					if old_year != cur_year:

						# Substitute the old year for the current year.
						find_line = ' %s, ' % old_year
						repl_line = ' %s, ' % cur_year
						line_ny = re.sub( find_line, repl_line, line )

						my_echo( "[updated ] %s" % filename )

						# Add the updated line to the running list.
						mod_file_lines += line_ny

					else:

						my_echo( "[up2date ] %s" % filename )

						# Add the unchanged line to the running list.
						mod_file_lines += line

				else:
					# Add the unchanged line to the running list.
					mod_file_lines += line

				# endif result

			# endfor

		else:

			# Don't go any further if we're only updating existing copyright
			# lines.
			if update_only:
				my_echo( "[nocrline] %s" % filename )
				continue

			num_file_lines = len( file_lines )

			# Iterate through the lines in the current file.
			for i in range( int(num_file_lines) ):

				line = file_lines[i]

				# Only look at the next line if we are not at the last line.
				if i < int(num_file_lines) - 1:
					line_next = file_lines[i+1]
				else:
					line_next = file_lines[i]

				# Try to match both the current line and the next line.
				result  = re.search( r'Copyright \(C\) ([0-9][0-9][0-9][0-9]), (.*)', line )
				resnext = re.search( r'Copyright \(C\) ([0-9][0-9][0-9][0-9]), (.*)', line_next )

				# Parse the results.
				if result:

					if resnext:

						# The current line matches but so does the next. Add the
						# current line unchanged to the running list.
						mod_file_lines += line

					else:

						# The current line matches but the next does not. Thus,
						# this branch only executes for the *last* copyright line
						# in the file.

						# Extract the year and organization from the matched
						# string.
						old_year = result.group(1)
						old_org  = result.group(2)

						# Set up search/replace strings to convert the current
						# line into one that serves as copyright for the_org.
						find_line = '%s, %s' % (old_year, old_org)
						repl_line = '%s, %s' % (cur_year, the_org)
						line_nyno = re.sub( find_line, repl_line, line )

						# Add the current line and then also insert our new
						# copyright line for the_org into the running list.
						mod_file_lines += line
						mod_file_lines += line_nyno

						my_echo( "[added   ] %s" % filename )

					# endif resnext

				else:

					# The current line does not match. Pass it through unchanged.
					mod_file_lines += line

				# endif result

			# endfor

		# endif has_org_cr

		if not dry_run:

			# Open the file for writing.
			f = open( filename, "w" )

			# Join the modified file lines into a single string.
			final_string = "".join( mod_file_lines )

			# Write the lines to the file.
			f.write( final_string )

			# Close the file.
			f.close()

		# endif not dry_run

	# Return from main().
	return 0




if __name__ == "__main__":
	main()
